import json
import unittest

from bs4 import BeautifulSoup
import datetime

from knowledge_repo import KnowledgeRepository, KnowledgePost
from knowledge_repo.app.app import db_session
from knowledge_repo.app.models import User, Post, Tag


class WebEditorPostTest(unittest.TestCase):
    """The test class for webeditor posts."""

    @classmethod
    def setUpClass(cls):
        """Set up the WebEditor Test Class.

        Create a webpost and a testuser
        """
        cls.repo = KnowledgeRepository.for_uri('sqlite:///:memory::test_posts', auto_create=True)
        cls.repo.config.editors = ['test_knowledge_editor']
        cls.app = cls.repo.get_app(config='tests/config_server.py')
        cls.app.config['AUTH_USERNAME_REQUEST_HEADER'] = 'user_header'
        cls.app.config['DEBUG'] = True
        cls.client = cls.app.test_client()

        cls.post_path = 'webeditor/misc/Test_Webeditor_Functionality'
        kp = KnowledgePost(cls.post_path)

        headers = {'title': 'Test Webeditor Functionality',
                   'authors': ['webeditor_test_user'],
                   'project': 'test_project',
                   'tldr': 'This is the test tldr for the webpost.',
                   'created_at': datetime.date.today(),
                   'updated_at': datetime.date.today()}
        md = 'Test Text'
        kp.write(md=md, headers=headers)
        cls.repo.add(kp)

        with cls.app.app_context():
            post = Post()
            db_session.add(post)
            post.update_metadata_from_kp(kp)
            db_session.commit()

            cls.post_id = post.id

    def test01_webeditor_route(self):
        """Test the /webposts route.

        This will need to test the security as well.
        The user 'Knowledge Default; is able to see all webposts,
        whereas 'Test User' should only see posts they've authored
        """

        with self.app.app_context():
            posts = db_session.query(Post).all()
            rv = self.client.get('/webposts', headers={'user_header': 'test_knowledge_editor'})

            assert rv.status == '200 OK'

            data = rv.data.decode('utf-8')
            soup = BeautifulSoup(data, 'html.parser')

            table_rows = soup.findAll("tr")

            # Subtract one for the header of the table
            assert len(table_rows) - 1 == len(posts)

            rv = self.client.get('/webposts', headers={'user_header': 'webeditor_test_user'})

            data = rv.data.decode('utf-8', 'html.parser')
            soup = BeautifulSoup(data, 'html.parser')
            table_rows = soup.findAll("tr")

            assert len(table_rows) - 1 == 1

    def test02_test_edit_form(self):
        """Test editing a webpost.

        Click on the edit button on the /webeditorpost route
        And ensure that the new route has the form fields
        filled out with the correct values
        """

        rv = self.client.get('/posteditor?post_id={}'.format(self.post_id))
        assert rv.status == "200 OK"
        data = rv.data.decode('utf-8')
        soup = BeautifulSoup(data, 'html.parser')

        with self.app.app_context():

            kp = self.repo.post(self.post_path)
            headers = kp.headers

            title = soup.findAll("input", {"id": "post_title"})[0]
            assert title['value'] == headers['title']

            # project = soup.findAll("input", {"id": "post_project"})[0]
            # assert project['value'] == post.project

            created_at = soup.findAll("input", {"id": "post_created_at"})[0]
            assert created_at['value'] == headers['created_at'].strftime('%Y-%m-%d')

            updated_at = soup.findAll("input", {"id": "post_updated_at"})[0]
            assert updated_at['value'] == headers['updated_at'].strftime('%Y-%m-%d')

            # TODO this now is populated by a "select"
            # author = soup.findAll("input", {"id": "post_author"})[0]
            # assert author['value'] == post.authors[0].username

            # feed_image = soup.findAll("input", {"id": "post_image_feed"})[0]
            # TODO(nikki_ray): For some reason the feed_image isn't being parsed
            # correctly
            # assert feed_image
            # assert feed_image['value']

            text = soup.findAll("textarea", {"id": "post_text"})[0]
            assert text.text.strip() == kp.read(headers=False).strip()

    def test03_test_save_button(self):
        """Test the save button on the webeditor form.

        Change the post text and make sure that
        the relevant field is updated in the db
        """
        new_post_text = "Here is the new post_text"

        with self.app.app_context():
            post = (db_session.query(Post)
                              .filter(Post.id == self.post_id)
                              .first())

            str_created_at = datetime.datetime.strftime(post.created_at, '%Y-%m-%d')
            str_updated_at = datetime.datetime.strftime(post.updated_at, '%Y-%m-%d')

            post_json_object = {
                'title': post.title,
                'project': post.project,
                'feed_image': "",  # TODO where do we store this?
                'tags': "",  # TODO pass tags
                'tldr': post.tldr,
                'status': post.status.value,
                'created_at': str_created_at,
                'updated_at': str_updated_at,
                'author': [auth.username for auth in post.authors],
                'markdown': new_post_text
            }

            url = "/save_post?post_id=" + str(post.id)
            rv = self.client.post(url,
                                  data=json.dumps(post_json_object),
                                  content_type='application/json')
            assert rv.status == "200 OK"

            kp = self.repo.post(self.post_path)

            assert kp.read(headers=False).strip() == new_post_text

    def test04_test_review_button(self):
        """Test the review functionality.

        Ensure that submitting a post for review changes the status
        And that going back to the form editor changes the button text
        And a review comment area is added
        """
        rv = self.client.post('/submit?post_id={}'.format(self.post_id),
                              data=json.dumps({'post_reviewers': 'test_post_reviewers'}),
                              content_type='application/json')

        assert rv.status == "200 OK"

        rv = self.client.get('/posteditor?post_id={}'.format(self.post_id))

        assert rv.status == "200 OK"

        data = rv.data.decode("utf-8")
        soup = BeautifulSoup(data, 'html.parser')

        comments_textarea = soup.findAll("textarea", {"id": "comment-text"})
        assert comments_textarea

        btn_in_review = soup.findAll("button", {"id": "btn_in_review"})
        assert btn_in_review and btn_in_review[0].text.strip() == "In Review Phase"

        with self.app.app_context():
            post = (db_session.query(Post)
                              .filter(Post.id == self.post_id)
                              .first())
            assert post.status == self.repo.PostStatus.SUBMITTED

    def test05_test_author_can_publish_button(self):
        """Test changing the permission for authors.

        Clicking the author_can_publish checkbox
        should change the relevant field in the db
        """
        url = "/author_publish?post_id=" + str(self.post_id)
        rv = self.client.post(url)

        assert rv.status == "200 OK"
        with self.app.app_context():
            post = (db_session.query(Post)
                              .filter(Post.id == self.post_id)
                              .first())
            kp = self.repo.post(post.path)
            assert kp.is_accepted is True

    def test06_test_publish_button(self):
        """Test the publish functionality.

        Test that clicking the publish button changes the status
        and the button text.

        Test that clicking it again changes the status back
        and the button text
        """
        # test that clicking the publish button changes the status
        rv = self.client.post('/publish_post?post_id={}'.format(self.post_id))

        assert rv.status == "200 OK"
        with self.app.app_context():
            post = (db_session.query(Post)
                              .filter(Post.id == self.post_id)
                              .first())
            assert post.status == self.repo.PostStatus.PUBLISHED
            rv = self.client.get('/posteditor?post_id={}'.format(self.post_id))

            assert rv.status == "200 OK"

            data = rv.data.decode("utf-8")
            soup = BeautifulSoup(data, 'html.parser')

            # go to the posteditor form to check the button
            btn_publish = soup.findAll("button", {"id": "btn_publish"})
            assert btn_publish and btn_publish[0].text.strip() == "Unpublish"

            rv = self.client.post('/unpublish_post?post_id={}'.format(self.post_id))
            assert rv.status == "200 OK"

            post = (db_session.query(Post)
                              .filter(Post.id == self.post_id)
                              .first())

            assert post.status == self.repo.PostStatus.UNPUBLISHED

            rv = self.client.get('/posteditor?post_id={}'.format(self.post_id))

            assert rv.status == "200 OK"

            data = rv.data.decode("utf-8")
            soup = BeautifulSoup(data, 'html.parser')

            # go to the posteditor form to check the button
            btn_publish = soup.findAll("button", {"id": "btn_publish"})
            assert btn_publish and btn_publish[0].text.strip() == "Publish"

    def test07_test_delete_button(self):
        """Test webpost deletion."""
        rv = self.client.get('/delete_post?post_id={}'.format(self.post_id))
        assert rv.status == "200 OK"

        with self.app.app_context():
            post = (db_session.query(Post)
                              .filter(Post.id == self.post_id)
                              .first())
            assert not post

    @classmethod
    def tearDownClass(cls):
        """Remove the user we created."""
        pass

if __name__ == '__main__':
    unittest.main()



import datetime

from ..postprocessor import KnowledgePostProcessor

REQUIRED_FIELD_TYPES = {
    'title': str,
    'authors': list,
    'created_at': datetime.datetime,
    'tldr': str,
    'tags': list
}

OPTIONAL_FIELD_TYPES = {
    'path': str,
    'updated_at': datetime.datetime
}


class FormatChecks(KnowledgePostProcessor):
    '''
    Use this to bootstrap your own KnowledgePostProcessor.
    '''
    _registry_keys = ['format_checks']

    @classmethod
    def process(cls, kp):
        headers = kp.headers
        for field, typ in REQUIRED_FIELD_TYPES.items():
            assert field in headers, "Required field `{}` missing from headers.".format(
                field)
            assert isinstance(headers[field], typ), "Value for field `{}` is of type {}, and needs to be of type {}.".format(
                field, type(headers[field]), typ)
        for field, typ in OPTIONAL_FIELD_TYPES.items():
            if field in headers:
                assert isinstance(headers[field], typ), "Value for field `{}` is of type {}, and needs to be of type {}.".format(
                    field, type(headers[field]), typ)



""" Define the routes that deal with tags

This includes:
  - /tag_pages
  - /batch_tags
  - /delete_tag
  - /delete_tag_posts
  - /edit_tag_description
  - /rename_tag
  - /remove_posts_tags
  - /toggle_tag_subscription
  - /tag_list
"""
from flask import request, render_template, Blueprint, g
from sqlalchemy import and_
import logging

from ..app import db_session
from ..models import PageView, Post, assoc_post_tag, Subscription, Tag
from ..utils.requests import from_request_get_feed_params
from ..utils.emails import send_subscription_email

blueprint = Blueprint('tag', __name__,
                      template_folder='../templates', static_folder='../static')


@blueprint.route('/batch_tags')
@PageView.logged
def render_batch_tags():
    """ Render the batch_tags page, which lists all the tags.

        From this view, editors can:
            - rename them
            - delete them
            - delete the tag from certain posts
            - add a tag description
    """
    # get the args
    # sort_by will either be "Tag" or "Number of Posts"
    sort_by = request.args.get('sort_by', '')
    sort_asc = request.args.get('sort_asc', '')
    sort_desc = not sort_asc
    feed_params = from_request_get_feed_params(request)

    # get all tags
    all_tags = db_session.query(Tag).all()
    tags_to_posts = {}
    nonzero_tags = []
    tags_to_description = {}
    for tag in all_tags:
        posts = tag.posts

        if not posts:
            db_session.delete(tag)
            db_session.commit()

        tags_to_posts[tag.id] = [(post.path, post.title) for post in posts]
        nonzero_tags.append(tag)
        # so that we can use the tag in the jinja template
        db_session.expunge(tag)

        tags_to_description[tag.id] = tag.description

    if sort_by:
        if sort_by == "Tag":
            def compare(x):
                return x.name.lower()
        elif sort_by == "Number_of_Posts":
            def compare(x):
                return len(tags_to_posts[x.id])
        else:
            def compare(x):
                return x
        nonzero_tags = sorted(nonzero_tags, key=compare, reverse=sort_desc)

    return render_template("batch_tags.html",
                           tags=nonzero_tags,
                           tags_to_posts=tags_to_posts,
                           tags_to_desc=tags_to_description,
                           feed_params=feed_params)


@blueprint.route('/delete_tag_post', methods=['GET', 'POST'])
@PageView.logged
def delete_tags_from_posts():
    """ Delete a tag from all the posts associated with it """
    tag_id = int(request.args.get('tag_id'))

    tag_obj = (db_session.query(Tag)
               .filter(Tag.id == tag_id)
               .first())

    delete_query = assoc_post_tag.delete().where(assoc_post_tag.c.tag_id == tag_id)
    db_session.execute(delete_query)
    db_session.delete(tag_obj)
    db_session.commit()

    return ""


@delete_tags_from_posts.object_extractor
def delete_tags_from_posts():
    tag_obj = Tag.query.filter(Tag.name == request.args.get('tag_name', '')).first()
    return {
        'id': tag_obj.id if tag_obj else None,
        'type': 'tag',
        'action': 'delete'
    }


@blueprint.route('/tag_pages')
@PageView.logged
def render_tag_pages():
    feed_params = from_request_get_feed_params(request)
    start = feed_params['start']
    num_results = feed_params['results']
    tag = request.args.get('tag', '')

    if tag[0] == '#':
        tag = tag[1:]

    tag_obj = (db_session.query(Tag)
               .filter(Tag.name == tag)
               .first())

    tag_description = tag_obj.description

    # Get subscription status
    subscribed = tag in feed_params["subscriptions"]

    # get all files with given tag
    tag_posts = tag_obj.posts
    posts = [post for post in tag_posts if post.is_published]

    posts = posts[start:min(start + num_results, len(posts))]

    post_stats = {post.path: {'all_views': post.view_count,
                              'distinct_views': post.view_user_count,
                              'total_likes': post.vote_count,
                              'total_comments': post.comment_count} for post in posts}

    author_to_count = {}
    for post in posts:
        authors = post.authors
        for author in authors:
            name = author.format_name
            if name in author_to_count:
                author_to_count[name] += 1
            else:
                author_to_count[name] = 1

    # get author with the max count
    max_count = 1
    max_author = ''
    for (author, count) in author_to_count.items():
        if count >= max_count:
            max_author = author
            max_count = count

    return render_template("tag_pages.html",
                           feed_params=feed_params,
                           full_tag=tag,
                           top_header=tag,
                           tag_description=tag_description,
                           tag_pocs=max_author,
                           posts=posts,
                           subscribed=subscribed,
                           post_stats=post_stats)


@render_tag_pages.object_extractor
def render_tag_pages():
    tag_obj = Tag.query.filter(Tag.name == request.args.get('tag', '')).first()
    return {
        'id': tag_obj.id if tag_obj is not None else None,
        'type': 'tag',
        'action': 'view'
    }


@blueprint.route('/edit_tag_description', methods=['POST'])
@PageView.logged
def edit_tag_desc():
    """ Edit the description of a tag. This is used in the tag_page """
    data = request.get_json()
    tag_id = int(data['tagId'])
    new_tag_desc = data['tagDesc']
    if new_tag_desc:
        tag = (db_session.query(Tag)
               .filter(Tag.id == tag_id)
               .first())
        tag._description = new_tag_desc
        db_session.commit()
    return ""


@edit_tag_desc.object_extractor
def edit_tag_desc():
    tag_obj = Tag.query.filter(Tag.id == request.get_json()['tagId']).first()
    return {
        'id': tag_obj.id if tag_obj is not None else None,
        'type': 'tag',
        'action': 'edit'
    }


@blueprint.route('/toggle_tag_subscription', methods=['GET', 'POST'])
@PageView.logged
def toggle_tag_subscription():
    """ Subscribe/Unsubscribe a user from a tag """
    try:
        # retrieve relevant tag and user args from request
        tag_name = request.args.get('tag_name', '')
        subscribe_action = request.args.get('subscribe_action', '')
        if subscribe_action not in ['unsubscribe', 'subscribe']:
            logging.warning("ERROR processing request")
            return ""

        # get the tag object so we can get the id
        tag_obj = (db_session.query(Tag)
                   .filter(and_(Tag.name == tag_name))
                   .first())
        # retrieve the current subscription object for a user and tag if exists
        subscription = None
        if tag_obj:
            subscription = db_session.query(Subscription).filter(
                and_(Subscription.object_type == 'tag',
                     Subscription.user_id == g.user.id,
                     Subscription.object_id == tag_obj.id)).first()
        if subscription and subscribe_action == 'unsubscribe':
            db_session.delete(subscription)
        elif not subscription and subscribe_action == 'subscribe':
            # otherwise, create new subscription
            subscription = Subscription(user_id=g.user.id, object_type='tag',
                                        object_id=tag_obj.id)
            db_session.add(subscription)
        else:
            logging.warning("ERROR processing request")
            return ""
        db_session.commit()
        return ""
    except:
        logging.warning("ERROR processing request")
        return ""


@toggle_tag_subscription.object_extractor
def toggle_tag_subscription():
    tag_obj = Tag(name=request.args.get('tag_name', ''))
    subscription = db_session.query(Subscription).filter(
        and_(Subscription.object_type == 'tag',
             Subscription.user_id == g.user.id,
             Subscription.object_id == tag_obj.id)
    ).first()
    return {
        'id': tag_obj.id if tag_obj is not None else None,
        'type': 'tag',
        'action': 'unsubscribe' if subscription else 'subscribe'
    }


@blueprint.route('/rename_tag', methods=['POST'])
@PageView.logged
def rename_tags_and_posts():
    """ Rename a tag
        This requires deleteing all the post-tag associations for the old tag
        and re-adding them for the new tag
    """
    data = request.get_json()
    old_tag_id = int(data['oldTagId'])
    new_tag = data['newTag']

    old_tag_obj = (db_session.query(Tag)
                   .filter(Tag.id == old_tag_id)
                   .first())

    new_tag_obj = (db_session.query(Tag)
                   .filter(Tag.name == new_tag)
                   .first())
    if not new_tag_obj:
        new_tag_obj = Tag(name=new_tag)
        db_session.commit()

    # add all the new associations in
    new_tag_post_ids_before = [p.id for p in new_tag_obj.posts]
    for p in old_tag_obj.posts:
        if p.id not in new_tag_post_ids_before:
            new_tag_obj.posts.append(p)
    db_session.commit()

    # delete the tag associations
    old_tag_obj.posts = []
    db_session.delete(old_tag_obj)
    db_session.commit()
    return ""


@rename_tags_and_posts.object_extractor
def rename_tags_and_posts():
    tag_id = int(request.get_json()['oldTagId'])
    return {
        'id': tag_id if tag_id else None,
        'type': 'tag',
        'action': 'rename'
    }


@blueprint.route('/remove_posts_tags', methods=['POST'])
@PageView.logged
def remove_posts_tags():
    """ Delete a tag from certain posts """
    data = request.get_json()
    tag_id = data['tagId']
    posts = data['posts']
    posts = [str(x) for x in posts]

    # for all the posts, query & remove that tag
    for post in posts:
        post_id = (db_session.query(Post).filter(Post.path == post).first()).id
        delete_query = assoc_post_tag.delete().where(
            and_(assoc_post_tag.c.tag_id == tag_id, assoc_post_tag.c.post_id == post_id))
        db_session.execute(delete_query)
    db_session.commit()
    return ""


@remove_posts_tags.object_extractor
def remove_posts_tags():
    return {
        'id': int(request.get_json()['tagId']),
        'type': 'tag',
        'action': 'dissociate'
    }


@blueprint.route('/tag_list', methods=['POST'])
@PageView.logged
def change_tags():
    """ Change the tags associated with a given post.
        This is called when someone clicks on the a knowledge post
        and changes the tag through the web ui
    """
    post_path = request.args.get('post_path', ' ')
    data = request.get_json()
    tags_string = data['tags']

    post = (db_session.query(Post)
            .filter(Post.path == post_path)
            .first())

    old_tags = post.tags
    tag_set = set(old_tags)
    tags_list = tags_string.split(",")
    post.tags = tags_list
    # post tag settr might take subset of tags_list to avoid dupes or whatever else
    new_tags = post.tags

    email_tags = [tag for tag in new_tags if tag not in tag_set]

    db_session.commit()

    for tag in email_tags:
        send_subscription_email(post, tag)

    return " "


@change_tags.object_extractor
def change_tags():
    post_path = request.args.get('post_path', '')
    return {
        'id': Post.query.filter(Post.path == post_path).first().id,
        'type': 'post',
        'action': 'tag'
    }



from __future__ import absolute_import
from builtins import next
from builtins import object
import os
import posixpath
import re
import collections
import logging
import datetime
import yaml
import mimetypes
import base64
import uuid

from .utils.encoding import encode, decode

logger = logging.getLogger(__name__)


SAMPLE_HEADER = """
---
title: "This is a Knowledge Post title, quoted so we can use special characters like ':'"
authors:
- sally_smarts
- wesly_wisdom
tags:
- knowledge
- example
created_at: 2016-06-29
updated_at: 2016-06-30
tldr: |
    You can write any markdown you want here (the '|' character makes this an escaped section)

    * bullet
    * bullet

    You can even write arbitrary html (html is valid markdown)
    <table>
    <tr>
    <th>hi</th>
    <th>I'm </th>
    </tr>
    <tr>
    <td>a</td>
    <td>table</td>
    </tr>
    </table>
---
"""


# Use OrderedDict representations in Yaml, so post headers remain in order
# Although this sets it for all uses of yaml in this Python instance,
# this shouldn't cause any problems.
def setup_yaml():
    _mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG

    def dict_representer(dumper, data):
        return dumper.represent_dict(data.items())

    def dict_constructor(loader, node):
        return collections.OrderedDict(loader.construct_pairs(node))

    yaml.add_representer(collections.OrderedDict, dict_representer)
    yaml.add_constructor(_mapping_tag, dict_constructor)


setup_yaml()


class ReferenceCache(object):

    def __init__(self, cache=None):
        self._cache = cache or {}

    def __setitem__(self, key, value):
        parents = posixpath.dirname(key).split('/')
        cache = self._cache
        for parent in parents:
            if parent:
                if parent not in cache:
                    cache[parent] = {}
                cache = cache[parent]
        cache[posixpath.basename(key)] = value

    def __getitem__(self, key):
        parents = posixpath.dirname(key).split('/')
        cache = self._cache
        for parent in parents:
            if parent:
                cache = cache[parent]
        return cache[posixpath.basename(key)]

    def __delitem__(self, key):
        parents = posixpath.dirname(key).split('/')
        cache = self._cache
        for parent in parents:
            if parent:
                cache = cache[parent]
        del cache[posixpath.basename(key)]

    def __getattr__(self, key):
        if key not in self._cache:
            raise AttributeError
        if isinstance(self._cache[key], dict):
            return ReferenceCache(self._cache[key])
        return self._cache[key]

    def keys(self):
        return list(self._cache.keys())

    def get(self, key, default=None):
        try:
            return self[key]
        except:
            return default

    def __contains__(self, key):
        try:
            parents = posixpath.dirname(key).split('/')
            cache = self._cache
            for parent in parents:
                if parent:
                    cache = cache[parent]
            return posixpath.basename(key) in cache
        except KeyError:
            return False

    def dir(self, parent=None, cache=None):
        if cache is None:
            cache = self._cache
            if parent:
                cache = cache.get(parent, {})
        for key in cache:
            if isinstance(cache[key], dict):
                for path in self.dir(parent=posixpath.join(parent or '', key), cache=cache[key]):
                    yield path
            else:
                yield posixpath.join(parent or '', key)


class KnowledgePost(object):
    '''
    A "Knowledge Post" is a (virtual) folder in which there is a 'knowledge.md' file,
    and potentially an 'images' and/or 'orig_src' folder. It is "virtual" in the sense
    that `KnowledgePost` objects store everything in memory, and use `KnowledgeRepository`
    object instances to interface with "physical" realisations of them. For example,
    in a GitKnowledgeRepository, the physical realisation is an actual folder; whereas
    in a database-backed KnowledgeRepository, on the virtual structure is maintained.
    '''

    def __init__(self, path=None, revision=None, repository=None):
        self.path = path
        self.revision = revision
        self.repository = repository
        self.__cache = ReferenceCache()  # Mapping of string filename to data

    @property
    def repository_uri(self):
        if (self.repository is not None):
            return self.repository._kp_repository_uri(self.path)

    @property
    def path(self):
        return self._path or self.headers.get('path')

    @path.setter
    def path(self, path):
        self._path = path

    @property
    def uuid(self):
        if not getattr(self, '_uuid', None):
            if self.repository is not None:
                self._uuid = self.repository._kp_uuid(self.path)  # Use repository UUID (even if None)
            else:
                self._uuid = str(uuid.uuid4())
        return self._uuid

    @uuid.setter
    def uuid(self, uuid):
        self._uuid = uuid

    # ------------ Reference cache methods ------------------------------------------
    def _read_ref(self, ref):
        if (ref not in self.__cache) and (self.repository is not None) and self.repository._kp_has_ref(self.path, ref):
            self.__cache[ref] = self.repository._kp_read_ref(self.path, ref)
        return self.__cache[ref]

    def _write_ref(self, ref, data):
        self.__cache[ref] = data

    def _drop_ref(self, ref):
        del self.__cache[ref]

    def _has_ref(self, ref):
        return (ref in self.__cache) or (self.repository is not None) and self.repository._kp_has_ref(self.path, ref)

    def _dir(self, parent=None):
        for ref in self.__cache.dir(parent=parent):
            if self.__cache[ref] is not None:
                yield ref
        if self.repository is not None:
            for ref in self.repository._kp_dir(self.path, parent=parent, revision=self.revision):
                if ref not in self.__cache:
                    yield ref

    # ---------- Knowledge Post Data Population -----------------------------

    def read(self, images=False, headers=True, body=True):
        if not (body or headers):
            md = ''
        else:
            md = decode(self._read_ref('knowledge.md'))
            mtch = re.match('^---\n[\s\S]+?---\n', md)
            if not mtch:
                raise ValueError("YAML header is missing. Please ensure that the top of your post has a header of the following form:\n" + SAMPLE_HEADER)
            if not headers:
                md = re.sub('^---\n[\s\S]+?---\n', '', md, count=1)
            if not body:
                md = mtch.group(0)
        if images:
            return md, self.read_images()
        return md

    @property
    def image_paths(self):
        return ['images/{}'.format(image_name) for image_name in self._dir(parent='images')]

    def read_image(self, name):
        return self._read_ref('image/' + name)

    def read_images(self):
        image_data = {}
        for image_path in self.image_paths:
            image_data[image_path] = self._read_ref(image_path)
        return image_data

    def read_src(self, ref):
        return self._read_ref('orig_src/' + ref)

    def write(self, md, headers=None, images={}):
        md = md.strip()
        if headers is not None:
            md = re.sub('^---\n[\s\S]+?---\n', '', md, count=1)
            md = '---\n' + \
                yaml.dump(headers, default_flow_style=False) + '---\n' + md
        md += '\n'

        self._write_ref('knowledge.md', encode(md))

        for image, data in list(images.items()):
            self._write_ref('images/' + image, data)

    def write_image(self, name, data):
        self._write_ref('images/' + name, data)

    def write_images(self, image_data={}):
        for name, data in list(self.image_data.items()):
            self.write_image(name, data)

    def write_src(self, name, data):
        self._write_ref('orig_src/' + name, encode(data))

    def add_srcfile(self, filename, name=None):
        if not name:
            name = os.path.basename(filename)
        with open(filename, 'rb') as f:
            self.write_src(name, f.read())

    # ------------- Knowledge Post Format ----------------------------------

    @property
    def headers(self):
        try:
            headers = next(yaml.load_all(self.read(body=False)))
        except StopIteration as e:
            raise ValueError("YAML header is missing. Please ensure that the top of your post has a header of the following form:\n" + SAMPLE_HEADER)
        except yaml.YAMLError as e:
            raise ValueError(
                "YAML header is incorrectly formatted or missing. The following information may be useful:\n{}\nIf you continue to have difficulties, try pasting your YAML header into an online parser such as http://yaml-online-parser.appspot.com/.".format(str(e)))
        for key, value in headers.copy().items():
            if isinstance(value, datetime.date):
                headers[key] = datetime.datetime.combine(value, datetime.time(0))
        return headers

    def update_headers(self, **headers):
        h = self.headers
        h.update(headers)
        self.write(self.read(), headers=h)

    @property
    def thumbnail_uri(self):
        image_matches = ExtractImages.find_images(self.read())
        if len(image_matches) == 0:
            return None

        i = 0
        image_uri = None
        while not image_uri:
            if i == len(image_matches):
                return None
            image_uri = image_matches[i]['src']
            i += 1

        if ':' not in image_uri:  # TODO: Check this works
            if not os.path.exists(image_uri):
                return None
            with open(self.feed_image[7:]) as f:
                data = base64.b64encode(f.read())
            image_mimetype = mimetypes.guess_type(self.feed_image)[0]
            if image_mimetype is not None:
                return 'data:{};base64,'.format(image_mimetype) + data

        return image_uri

    def is_valid(self):
        if not self._has_ref('knowledge.md'):
            return False
        try:
            FormatChecks.process(self)
        except:
            return False
        return True

    @property
    def status(self):
        if self.repository is not None:
            return self.repository._kp_status(self.path, revision=self.revision)

    @property
    def is_published(self):
        return self.status == self.repository.PostStatus.PUBLISHED

    @property
    def is_accepted(self):
        return self.status in [self.repository.PostStatus.UNPUBLISHED, self.repository.PostStatus.PUBLISHED]

    @property
    def web_uri(self):
        if self.repository is not None:
            return self.repository._kp_web_uri(self.path)

    # Conversion/Import/Export methods
    @classmethod
    def from_file(cls, filename, src_paths=[], format=None, postprocessors=None, **opts):
        kp = KnowledgePostConverter.for_file(cls(), filename, format=format, postprocessors=postprocessors).from_file(filename, **opts)
        if src_paths:
            for src_path in src_paths:
                kp.add_srcfile(src_path)
        return kp

    @classmethod
    def from_string(cls, string, src_strings={}, format=None, postprocessors=None, **opts):
        kp = KnowledgePostConverter.for_format(cls(), format=format, postprocessors=postprocessors).from_string(string, ** opts)
        if src_strings:
            for src_name, data in list(src_strings.items()):
                kp.write_src(src_name, data)
        return kp

    def to_file(self, filename, format=None, **opts):
        return KnowledgePostConverter.for_file(self, filename, format=format).to_file(filename, **opts)

    def to_string(self, format, **opts):
        return KnowledgePostConverter.for_format(self, format).to_string(**opts)

from .converter import KnowledgePostConverter  # noqa
from .postprocessors.format_checks import FormatChecks  # noqa
from .postprocessors.extract_images import ExtractImages  # noqa



import unittest
from bs4 import BeautifulSoup
from sqlalchemy import and_

from knowledge_repo import KnowledgeRepository
from knowledge_repo.app.models import User, Vote, Post
from knowledge_repo.app.app import db_session


class FavoriteTest(unittest.TestCase):

    def setUp(self):
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.repo_app = self.repo.get_app(config='tests/config_server.py')
        self.app = self.repo_app.test_client()
        self.headers = {}

        self.knowledge_username = 'favorite_test_user'
        username_request_header = self.repo_app.config.get(
            "AUTH_USERNAME_REQUEST_HEADER")
        self.headers = {
            username_request_header: self.knowledge_username
        }

    def test_like_and_unlike_a_post(self):
        """
        Like and then unlike a post
        """
        with self.repo_app.app_context():
            post = (db_session.query(Post)
                    .filter(Post.is_published)
                    .first())

            old_likes = (db_session.query(Vote)
                         .filter(Vote.object_id == post.id)
                         .filter(Vote.object_type == 'post')
                         .all())
            rv = self.app.get("/like?post_id=" + str(post.id), headers=self.headers)

            assert rv.status == '200 OK', post.path + rv.status

            new_likes = (db_session.query(Vote)
                         .filter(Vote.object_id == post.id)
                         .filter(Vote.object_type == 'post')
                         .all())

            assert len(old_likes) + 1 == len(new_likes)

            # assert that if you re-like the page, the number of likes doesn't
            # change
            rv = self.app.get("/like?post_id=" + str(post.id), headers=self.headers)

            assert rv.status == '200 OK'

            like_again = (db_session.query(Vote)
                          .filter(Vote.object_id == post.id)
                          .filter(Vote.object_type == 'post')
                          .all())
            assert len(like_again) == len(new_likes)

            """
            Let's unlike it again
            """
            old_likes = (db_session.query(Vote)
                         .filter(Vote.object_id == post.id)
                         .filter(Vote.object_type == 'post')
                         .all())
            rv = self.app.get("/unlike?post_id=" + str(post.id), headers=self.headers)

            assert rv.status == '200 OK'

            new_likes = (db_session.query(Vote)
                         .filter(Vote.object_id == post.id)
                         .filter(Vote.object_type == 'post')
                         .all())
            assert len(new_likes) == len(old_likes) - 1

    def test_favorites(self):
        """
        test the favorites route, and make sure the number of posts = number of votes with that user_id
        """
        with self.repo_app.app_context():
            rv = self.app.get("/favorites", headers=self.headers)
            assert rv.status == "200 OK"
            data = rv.data.decode('utf-8')
            soup = BeautifulSoup(data, 'html.parser')
            all_posts = soup.findAll(
                'div', {'class': 'row row-space-4 panel feed-post'})

            user = (db_session.query(User)
                    .filter(User.username == self.knowledge_username)
                    .first())
            votes = (db_session.query(Vote)
                     .filter(Vote.user_id == user.id)
                     .all())
            assert len(votes) == len(all_posts)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
import os
import imp
import logging
import traceback

from flask import Flask, current_app, render_template, g, request
from flask_mail import Mail
from flask_migrate import Migrate
from alembic import command
from alembic.migration import MigrationContext

from .proxies import db_session, current_repo
from .index import update_index
from .models import db as sqlalchemy_db, Post, User
from . import routes

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class KnowledgeFlask(Flask):

    def __init__(self, repo, db_uri=None, debug=None, config=None, **kwargs):
        Flask.__init__(self, __name__,
                       template_folder='templates',
                       static_folder='static')

        # Load configuration
        if config:
            if isinstance(config, str):
                config = imp.load_source('knowledge_server_config', os.path.abspath(config))
            self.config.from_object(config)
        self.config.update(kwargs)

        if hasattr(config, 'prepare_repo'):
            repo = config.prepare_repo(repo)
        self.repository = repo

        # Lift debug mode from underlying KnowledgeRepository object if not specified
        self.config['DEBUG'] = debug if debug is not None else repo.config.debug

        # Configure database
        db_uri = db_uri or self.config.get('SQLALCHEMY_DATABASE_URI') or 'sqlite:///:memory:'
        logger.debug("Using database: {}".format(db_uri))
        self.config['SQLALCHEMY_DATABASE_URI'] = db_uri
        self.config['SQLALCHEMY_ECHO'] = repo.config.debug
        self.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

        # Register database schema with flask app
        sqlalchemy_db.init_app(self)

        # Set up database migration information
        # Registers Migrate plugin in self.extensions['migrate']
        Migrate(self, self.db)

        # Try to create the database if it does not already exist
        # Existence is determined by whether there is an existing alembic migration revision
        db_auto_create = self.config.get('DB_AUTO_CREATE', True)
        db_auto_upgrade = self.config.get('DB_AUTO_UPGRADE', True)
        if db_auto_create and self.db_revision is None:
            self.db_init()
        elif db_auto_upgrade:
            self.db_upgrade()

        # Add mail object if configuration is supplied
        if self.config.get('MAIL_SERVER', None) is not None:
            self.config['mail'] = Mail(self)

        # Set config defaults if not included
        server_config_defaults = {'SERVER_NAME': 'localhost',
                                  'plugins': []}
        for k, v in server_config_defaults.items():
            self.config[k] = self.config.get(k, v)

        # Register routes to be served
        self.register_blueprint(routes.render.blueprint)
        self.register_blueprint(routes.health.blueprint)
        self.register_blueprint(routes.index.blueprint)
        self.register_blueprint(routes.tags.blueprint)
        self.register_blueprint(routes.vote.blueprint)
        self.register_blueprint(routes.comment.blueprint)
        self.register_blueprint(routes.stats.blueprint)
        self.register_blueprint(routes.web_editor.blueprint)

        # Register error handler
        @self.errorhandler(500)
        def show_traceback(self):
            """ If LOCAL MODE: show the stack trace on a server error
                otherwise show a nice error template to the users
            """
            if current_app.config.get("DEBUG"):
                return render_template('traceback.html', info=traceback.format_exc()), 500
            else:
                return render_template('error.html')

        @self.before_first_request
        def update_index_and_typeahead_data():
            """ Function to call all of the steps we want to run recurrently
                Ideally, these functions would be in a background job, but
                running them on each user's first request is a workable
                solution for now.

                Putting these functions in a function wrapper also helps
                because before_first_request decorated functions cannot be
                imported, so this style allows the functions to be used
                across the app in addition to being run at first.
            """

            if not current_app.config.get('REPOSITORY_INDEXING_ENABLED', True):
                return

            typeahead_data = {}

            def update_typeahead_data(post):
                """ Create the typeahead entry for a given knowledge post """
                if post.authors and post.title and post.path:
                    authors_str = [author.format_name for author in post.authors]
                    typeahead_entry = {'author': authors_str,
                                       'title': str(post.title),
                                       'path': str(post.path),
                                       'updated_at': str(post.updated_at)}
                    typeahead_data[str(post.title)] = typeahead_entry

            update_index()
            posts = (db_session.query(Post)
                     .filter(Post.is_published)
                     .all())

            for post in posts:
                update_typeahead_data(post)

            current_app.config['typeahead_data'] = typeahead_data

        @self.before_request
        def set_user_information():
            if not request.path.startswith('/static'):
                username = current_app.config.get('AUTH_USERNAME_DEFAULT') or 'knowledge_default'
                if current_app.config.get('AUTH_USERNAME_REQUEST_HEADER') is not None:
                    username = request.headers.get(current_app.config.get(
                        'AUTH_USERNAME_REQUEST_HEADER'), username)
                g.user = User(username=username)
                if g.user.id is None:
                    db_session.commit()

        @self.before_request
        def open_repository_session():
            if not request.path.startswith('/static'):
                current_repo.session_begin()

        @self.after_request
        def close_repository_session(response):
            if not request.path.startswith('/static'):
                current_repo.session_end()
            return response

        @self.context_processor
        def webediting_enabled():
            # TODO: Link this more to KnowledgeRepository capability and
            # configuration rather than a specific name.
            return {"webeditor_enabled": 'webposts' in current_repo.uris}

    @property
    def repository(self):
        return getattr(self, '_repository')

    @repository.setter
    def repository(self, repo):
        self._repository = repo

    @property
    def db(self):
        return sqlalchemy_db

    @property
    def _alembic_config(self):
        if not hasattr(self, 'extensions') or 'migrate' not in self.extensions:
            raise RuntimeError("KnowledgeApp has not yet been configured. Please instantiate it via `get_app_for_repo`.")
        migrations_path = os.path.join(os.path.dirname(__file__), "migrations")
        return self.extensions['migrate'].migrate.get_config(migrations_path)

    def db_init(self):
        with self.app_context():
            # Create all tables
            sqlalchemy_db.create_all()

            # Stamp table as being current
            command.stamp(self._alembic_config, "head")

    @property
    def db_revision(self):
        with self.app_context():
            conn = self.db.engine.connect()

            context = MigrationContext.configure(conn)
            return context.get_current_revision()

    def db_upgrade(self):
        with self.app_context():
            command.upgrade(self._alembic_config, "head")

    def db_migrate(self, message, autogenerate=True):
        with self.app_context():
            command.revision(self._alembic_config, message=message, autogenerate=autogenerate)



import os

from ..converter import KnowledgePostConverter
from .._version import __optional_dependencies__

TEMPLATE = '''
{%- extends 'markdown.tpl' -%}

{%- block input -%}
{%- if cell['metadata'].get('slideshow',{}).get('slide_type','') == 'skip' -%}
{%- else %}
```python
{{ cell.source }}
```
{%- endif %}
{%- endblock input -%}

{%- block data_priority scoped %}
{{ super() }}
{%- endblock %}


{# remove stderr output #}
{%- block stream scoped -%}
    {%- if output.name == 'stdout' -%}
        {{ super () }}
    {%- elif output.name == 'stderr' -%}
    {%- endif -%}
{%- endblock stream -%}
'''


class IpynbFormat(KnowledgePostConverter):
    _registry_keys = ['ipynb']

    @property
    def dependencies(self):
        return __optional_dependencies__['ipynb']

    def from_file(self, filename):
        import nbformat
        from nbconvert import MarkdownExporter
        from jinja2 import DictLoader
        from traitlets.config import Config

        c = Config()
        # c.ExtractOutputPreprocessor.extract_output_types = set()
        c.ExtractOutputPreprocessor.output_filename_template = 'images/{unique_key}_{cell_index}_{index}{extension}'

        nb = nbformat.read(filename, as_version=4)

        dl = DictLoader({'full.tpl': TEMPLATE})
        md_exporter = MarkdownExporter(config=c, extra_loaders=[
                                       dl], template_file='full.tpl')
        (body, resources) = md_exporter.from_notebook_node(nb)

        self.kp.write(body, images={name.split(
            'images/')[1]: data for name, data in resources.get('outputs', {}).items()})

        # Add cleaned ipynb file
        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                cell['outputs'] = []  # remove output data
                cell['execution_count'] = None  # reset to not executed
        self.kp.write_src(os.path.basename(filename), nbformat.writes(nb))



"""
Check the health of the app.

This view defines routes that allow us to check
the health of the server.
Ultimately, we can do something more complex
"""
from flask import Blueprint

blueprint = Blueprint('health', __name__,
                      template_folder='../templates', static_folder='../static')


@blueprint.route('/health')
def health():
    """Return OK when a request to the health view is received."""
    return "OK"


@blueprint.route('/ping')
def ping():
    """Return OK when a request to the ping view is received."""
    return "OK"



import os
from setuptools import setup, find_packages
from distutils import log
from distutils.command.install_scripts import install_scripts

# Extract version info from library
version_info = {}
with open('knowledge_repo/_version.py') as version_file:
    exec(version_file.read(), version_info)

# Batch file template for Windows (based on https://github.com/matthew-brett/myscripter/blob/master/setup.py)
BATCHFILE_TEMPLATE = r"""
@echo off
REM Wrapper around {script_name} to execute script in Windows using the interpreter specified with the hashbang
set dirname=%~dp0
set wrapped_script="%dirname%{script_name}"
set /p line1=<%wrapped_script%
if "%line1:~0,2%" == "#!" (goto :execute)
echo First line of %wrapped_script% does not start with "#!"
exit /b 1
:execute
set py_exe=%line1:~2%
call "%py_exe%" %wrapped_script% %*
"""


# Custom script installer for Windows (based on https://github.com/matthew-brett/myscripter/blob/master/setup.py)
class install_scripts_windows_wrapper(install_scripts):
    def run(self):
        install_scripts.run(self)
        if not os.name == "nt":
            return
        for script_path in self.get_outputs():
            # If we can find an executable name in the #! top line of the script
            # file, make .bat wrapper for script.
            with open(script_path) as fobj:
                first_line = fobj.readline()
            if not (first_line.startswith('#!') and 'python' in first_line.lower()):
                log.info("Script does not appear to be a python executable. Skipping creation of batchfile wrapper")
                continue
            script_dirname, script_basename = os.path.split(script_path)
            script_name, _ = os.path.splitext(script_basename)
            batchfile_path = os.path.join(script_dirname, script_name + '.bat')
            log.info("Making batchfile wrapper at {} (for {})".format(batchfile_path, script_path))

            batchfile_content = BATCHFILE_TEMPLATE.format(script_name=script_name)

            if self.dry_run:
                continue
            with open(batchfile_path, 'w') as fobj:
                fobj.write(batchfile_content)


setup(
    name='knowledge_repo',
    description=(
        " A workflow for contributing company knowledge, in the form "
        " of RMarkdowns, iPythons, and Markdowns, rendered and organized"
        " to magnify research impact across teams and time "),
    version=version_info['__version__'].split('_')[0],  # remove git revision if present
    author=version_info['__author__'],
    author_email=version_info['__author_email__'],
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,  # See included paths in MANIFEST.in
    scripts=['scripts/knowledge_repo'],
    install_requires=version_info['__dependencies__'],
    extras_require=version_info['__optional_dependencies__'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    cmdclass={'install_scripts': install_scripts_windows_wrapper}
)



import os
import sys
import logging

from future.utils import raise_with_traceback

__all__ = ['encode', 'decode']

logger = logging.getLogger(__name__)


def encode(data, encoding='utf-8'):
    assert encoding == 'utf-8', "Only UTF-8 encoding is currently supported."
    # Check if data is alreqdy encoded (for the purposes of unicode only
    # If not, convert to a string if necessary, and then encode as utf-8 bytes
    if sys.version_info.major == 2:
        if not isinstance(data, basestring):
            data = unicode(data)
        elif not isinstance(data, unicode):
            return data  # UTF-8 data is already encoded
    else:
        if isinstance(data, bytes):
            return data  # UTF-8 data is already encoded
        elif not isinstance(data, str):
            data = str(data)
    # Encode UTF-8 data
    if encoding is not None:
        try:
            data = data.encode(encoding)
        except Exception as e:
            if os.environ.get('DEBUG'):
                raise_with_traceback(e)
            logger.warning("An encoding error has occurred... continuing anyway. To capture these errors, rerun the current command prefixed with `DEBUG=1 `.")
            data = data.encode(encoding, errors='ignore')
    return data


def decode(data, encoding='utf-8'):
    assert encoding == 'utf-8', "Only UTF-8 encoding is currently supported."
    if encoding is not None:
        try:
            data = data.decode(encoding)
        except Exception as e:
            if os.environ.get('DEBUG'):
                raise_with_traceback(e)
            logger.warning("An decoding error has occurred... continuing anyway. To capture these errors, rerun the current command prefixed with `DEBUG=1 `.")
            data = data.decode(encoding, errors='ignore')
    return data



"""Parses request information.

Functions include:
    - check_user
    - from_request_get_user_info
    - from_request_get_feed_params
"""
import inflection
from flask import g

from ..app import db_session
from ..models import User


def from_request_get_feed_params(request):
    """Given the request, return an object that contains the parameters.

    :param request: The request obj
    :type request: Flask request
    :return: Select parameters passed in through headers or the url
    :rtype: object
    """

    feed_params = {}

    feed_params["filters"] = request.args.get('filters')
    feed_params["authors"] = request.args.get('authors')
    feed_params["start"] = int(request.args.get('start', 0))
    feed_params["results"] = int(request.args.get('results', 10))
    feed_params["sort_by"] = inflection.underscore(
        request.args.get('sort_by', 'updated_at'))
    feed_params["sort_desc"] = not bool(request.args.get('sort_asc', ''))
    username, user_id = g.user.username, g.user.id
    feed_params["username"] = username
    feed_params["user_id"] = user_id

    user_obj = (db_session.query(User)
                          .filter(User.id == user_id)
                          .first())

    if user_obj:
        feed_params["subscriptions"] = user_obj.get_subscriptions
    return feed_params



import unittest
from bs4 import BeautifulSoup

from knowledge_repo import KnowledgeRepository
from knowledge_repo.app.app import db_session
from knowledge_repo.app.models import PageView, Comment, Vote
from sqlalchemy import func


class FeedTest(unittest.TestCase):

    def setUp(self):
        self.repo = KnowledgeRepository.for_uri('tests/test_repo', auto_create=True)
        self.repo_app = self.repo.get_app(config='tests/config_server.py')
        self.app = self.repo_app.test_client()
        self.headers = {}
        self.data = []

    def test01_ten_posts(self):
        """
        test to make sure we have 10 posts
        test helper functions to make sure that the likes, pageviews, comments
        and metadata are correct
        """
        rv = self.app.get("/feed", headers=self.headers)
        assert (rv.status == "200 OK")
        data = rv.data.decode('utf-8')
        soup = BeautifulSoup(data, 'html.parser')
        all_posts = soup.findAll('div', {'class': 'row row-space-4 panel feed-post'})
        self.data = all_posts

    def test02_test_link_to_post(self):
        """
        For the posts, test to make sure that a link exists
        and that hitting that link gives a 200
        """
        posts = self.data

        for post in posts:
            title_link = post.findAll("a")[0]
            title_href = title_link['href']
            title_text = title_link.string

            assert(len(title_href))
            assert(self.app.get(title_href, headers=self.headers).status == "200 OK")

    def test03_test_post_stats(self):
        """
        Test post_stats
        """
        posts = self.data

        for post in posts:
            title_link = post.findAll("a")[0]
            title_href = title_link['href']
            find_equal = title_href.find("=")
            page_id = title_href[find_equal + 1:]

            # get pageviews
            # TODO: refine this for latest pageview model
            pageviews = post.findAll(
                "i", id=lambda x: x and x.startswith('tooltip-pageviews'))
            assert (len(pageviews) == 1)
            pageview_value = int(pageviews[0].text)
            pageview_query = (db_session.query(func.count(PageView.id))
                              .filter(PageView.page == page_id)
                              .scalar())
            assert (pageview_query == pageview_value + 1)
            # get likes
            likes = post.findAll(
                "i", id=lambda x: x and x.startswith('tooltip-likes'))
            assert (len(likes) == 1)

            likes_value = int(likes[0].text)
            likes_query = (db_session.query(Vote)
                           .filter(Vote.page == page_id)
                           .all())
            assert (likes_value == len(likes_query))

            # get comments
            comments = post.findAll(
                "i", id=lambda x: x and x.startswith('tooltip-comments'))
            assert (len(comments) == 1)

            comments_values = int(comments[0].text)
            comments_query = (db_session.query(Comment)
                              .filter(Comment.post_id == page_id)
                              .all())
            assert (comments_values == len(comments_query))

    def test04_test_post_metadata(self):
        """
        Test post metadata
        """
        posts = self.data

        for post in posts:
            author = post.findAll('li', {'class': 'feed-author'})
            assert author

            date = post.findAll('li', {'class': 'feed-date'})
            assert date

            tags = post.findAll('li', {'class': 'feed-tags'})
            assert tags

            image = post.findAll('img')
            assert image

    def test05_test_bogus_filter(self):
        """
        Test that a bogus filter returns 0 posts
        """
        url = '/feed?filters=bogus%20filter'
        rv = self.app.get(url, headers=self.headers)

        assert rv.status == '200 OK'

        data = rv.data.decode('utf-8')
        soup = BeautifulSoup(data, 'html.parser')
        all_posts = soup.findAll('div', {'class': 'row row-space-4 panel feed-post'})
        assert not all_posts

    def test06_real_filter(self):
        """
        Test that search filter "template" returns the three posts
        """
        url = '/feed?filters=template'
        rv = self.app.get(url, headers=self.headers)

        assert rv.status == '200 OK'

        data = rv.data.decode('utf-8')
        soup = BeautifulSoup(data, 'html.parser')
        all_posts = soup.findAll('div', {'class': 'row row-space-4 panel feed-post'})
        assert len(all_posts) == 3, "Expected 3 posts, found {}".format(len(all_posts))

if __name__ == '__main__':
    unittest.main()



from ..postprocessor import KnowledgePostProcessor


class StubPostprocessor(KnowledgePostProcessor):
    '''
    Use this to bootstrap your own KnowledgePostProcessor.
    '''
    _registry_keys = None

    @classmethod
    def process(cls, kp):
        pass



import posixpath

# This file is only needed for setting up a dedicated server configuration
# with support for extracting username information and emailing users.

SERVER_NAME = 'localhost'

# ---------------------------------------------------
# Database configuration
# ---------------------------------------------------
SQLALCHEMY_DATABASE_URI = 'sqlite:///' + posixpath.join(posixpath.dirname(__file__), 'knowledge.db')

# Should the database tables be automatically created
DB_AUTO_CREATE = True

# Should the database be automatically migrated when updates exist
# Note: This is True by default if this configuration is not applied,
# under the assumption that if you are concerned about this file
# you are probably interested in minimising risk to stability and handling
# database upgrades manually. Manual database migrations can be
# performed using `knowledge_repo --repo <> db_upgrade ...`.
DB_AUTO_UPGRADE = False


# ---------------------------------------------------
# Repository configuration
# ---------------------------------------------------
# You may specify a function `prepare_repo` which configures
# the repository upon which this server is running. This
# takes place after the repository has been instantiated
# and before the server is able to serve requests. It is
# possible to do anything to the repository, including
# substituting the repository for another one.
# By default, repositories manage their own configurations,
# but this can be risky as they may run arbitrary python code,
# which opens a vector for malicious users to compromise
# the server. If you want to avoid this risk, pass
# the '--safe' (TODO!) option to `knowledge_repo` config and
# manually configure the repository here.
# For example, if your server instance is sitting atop
# a meta-repository, it may make sense to update the meta-repository
# configuration with that of one of its children.
def prepare_repo(repo):
    return repo

# ---------------------------------------------------
# Authentication Headers
# ---------------------------------------------------
# the username request header should allow you to get
# a person's username - this is used for pageview tracking,
# likes, favorites, etc. As a fallback, DEFAULT_USER is used.
AUTH_USERNAME_REQUEST_HEADER = 'test_user_request_header'
AUTH_USERNAME_DEFAULT = 'knowledge_default'

# the group request header should allow you to get a
# user's name. This is then used below, where we can
# map specific users to specific projects for the
# webeditor posts
AUTH_GROUP_REQUEST_HEADER = 'test_group_request_header'
AUTH_GROUP_DEFAULT = ''

USER_GROUP_TO_PROJECT = {}  # TODO: Deprecate?

# ---------------------------------------------------
# Flask Mail Configuration
# Refer to https://pythonhosted.org/flask-mail/
# Unless specified, upstream defaults are used as indicated
# provided that MAIL_SERVER is defined.
# ---------------------------------------------------
TESTING = True
MAIL_SERVER = 'localhost'  # default = 'localhost'
# MAIL_PORT = 25  # default = 25
# MAIL_USE_TLS = False  # default = False
# MAIL_USE_SSL = False  # default = False
# MAIL_DEBUG = False  # default = app.debug
# MAIL_USERNAME = None  # default = None
# MAIL_PASSWORD = None  # default = None
MAIL_DEFAULT_SENDER = "knowledge_editors"  # default = None
# MAIL_MAX_EMAILS = None  # default = None
# MAIL_SUPPRESS_SEND = False  # default = app.testing
# MAIL_ASCII_ATTACHMENTS = False  # default = False

# --------------------------------------------------
# Web Editor Configuration
# --------------------------------------------------
# The web editor can be limited to editing posts under
# a limited set of parent directories by setting
# WEB_EDITOR_PREFIXES to a list of supported path prefixes.
# e.g. ['webposts', 'projects']
WEB_EDITOR_PREFIXES = None



from __future__ import absolute_import
from __future__ import print_function

from builtins import object
import sys
import os
import posixpath
from abc import abstractmethod, abstractproperty
import datetime
from collections import OrderedDict
from enum import Enum
import uuid

from . import config_defaults
from .post import KnowledgePost
from .config import KnowledgeRepositoryConfig
from .postprocessor import KnowledgePostProcessor
from .utils.registry import SubclassRegisteringABCMeta
from future.utils import with_metaclass

if sys.version_info.major > 2:
    from urllib.parse import urlparse
else:
    from urlparse import urlparse


class KnowledgeRepository(with_metaclass(SubclassRegisteringABCMeta, object)):
    _registry_keys = None

    class PostStatus(Enum):
        '''
        Do not store these values in a datastore, as they may change
        from release to release. These keys should only be used to compare with
        the output of a KnowledgeRepository.
        '''
        DRAFT = 0  # Post is still being written and not yet submitted
        SUBMITTED = 1  # Post is submitted and waiting for review
        UNPUBLISHED = 2  # Post is approved to publish, but not published
        PUBLISHED = 3  # Post is published and visible on /feed

    @classmethod
    def for_uri(cls, uri, *args, **kwargs):
        if isinstance(uri, dict):
            return cls.for_uris(uri)
        scheme = urlparse(uri).scheme
        return cls._get_subclass_for(scheme)(uri, *args, **kwargs)

    @classmethod
    def for_uris(cls, uri):
        # Import this within this method so as not to cause import resolution problems
        from .repositories.meta import MetaKnowledgeRepository
        if isinstance(uri, str):
            uris = {'': uri}
        else:
            uris = uri

        krs = {name: cls.for_uri(uri) for name, uri in list(uris.items())}
        return MetaKnowledgeRepository(krs)

    @classmethod
    def create_for_uri(cls, uri, **kwargs):
        if isinstance(uri, dict):
            return cls.for_uris(uri)
        scheme = urlparse(uri).scheme
        return cls._get_subclass_for(scheme).create(uri, **kwargs)

    @classmethod
    def create(cls, uri, **kwargs):
        raise NotImplementedError

    def __init__(self, uri, debug=False, **kwargs):
        self.uri = uri
        self.config = KnowledgeRepositoryConfig()
        self.config.debug = debug
        self.config.update_defaults(config_defaults)
        self.init(**kwargs)

    def init(self):
        pass

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, config):
        assert isinstance(config, KnowledgeRepositoryConfig), "`config` should be a `KnowledgeRepositoryConfig` instance."
        self._config = config

    @property
    def uris(self):
        # This translates KnowledgeRepository.uri to a consistent format
        # across all KnowledgeRepository instances: a dictionary of form
        # {<mountpoint>: <uri>}
        # It assumes that self.uri is either a string or a dictionary mapping
        # of form:
        # {<mountpoint>: <KnowledgeRepositoryInstance>}
        if isinstance(self.uri, str):
            return {'': self.uri}
        elif isinstance(self.uri, dict):
            uri_dict = {}

            def add_uris(uri_dict, uris, parent=''):
                assert isinstance(uris, dict)
                for mountpoint, uri in uris.items():
                    if isinstance(uri, (str, KnowledgeRepository)):
                        uri_dict[posixpath.join(parent, mountpoint)] = uri if isinstance(uri, str) else uri.uri
                    elif isinstance(uri, dict):
                        add_uris(uri_dict, uri, parent=posixpath.join(parent, mountpoint))
                    else:
                        raise ValueError("Unrecognised uri: {}".format(uri))
            add_uris(uri_dict, self.uri)
            return uri_dict
        raise ValueError("Unrecognised KnowledgeRepository.uri: {}".format(self.uri))

    # ------------- Repository actions / state ------------------------------------

    def session_begin(self):
        pass

    def session_end(self):
        pass

    @abstractproperty
    def revision(self):
        raise NotImplementedError

    def update(self):
        pass

    @abstractproperty
    def status(self):
        raise NotImplementedError

    @abstractproperty
    def status_message(self):
        raise NotImplementedError

    def set_active_draft(self, path):
        pass

    # -------------- Post retrieval methods --------------------------------------

    def post(self, path, revision=None):
        path = self._kp_path(path)
        if not self.has_post(path, revision=revision) and path in self.config.aliases:
            path = self.config.aliases[path]
            if path in self.config.alias:
                raise ValueError("Alias cycle detected.")
        assert self.has_post(path, revision=revision), "{} does not have a post for path '{}'.".format(self.__class__.__name__, path)
        return KnowledgePost(path=path, repository=self, revision=revision or self._kp_get_revision(path))

    def dir(self, prefix=None, status=None):
        if prefix is None or isinstance(prefix, str):
            prefixes = [prefix]
        else:
            prefixes = prefix
        assert all([prefix is None or isinstance(prefix, str) for prefix in prefixes]), "All path prefixes must be strings."
        prefixes = [prefix if prefix is None else posixpath.relpath(prefix, '/') for prefix in prefixes]
        if isinstance(status, str):
            if status == 'all':
                status = [self.PostStatus.DRAFT, self.PostStatus.SUBMITTED, self.PostStatus.PUBLISHED, self.PostStatus.UNPUBLISHED]
            else:
                raise ValueError('Status alias `{}` not recognised.'.format(status))
        if status is not None and not isinstance(status, list):
            status = [status]
        elif status is None:
            status = [self.PostStatus.PUBLISHED]

        # Use old syntax for "yielding from" to maintain support for python 2
        for prefix in prefixes:
            for path in self._dir(prefix=prefix, statuses=status):
                yield path

    @abstractmethod
    def _dir(self, prefix, statuses):
        raise NotImplementedError

    def has_post(self, path, revision=None):
        return self._kp_exists(self._kp_path(path), revision=revision)

    def post_status(self, path, revision=None, detailed=False):
        return self._kp_status(self._kp_path(path), revision=revision, detailed=detailed)

    def post_statuses(self, paths, detailed=False):
        return OrderedDict([(path, self.post_status(path, detailed=detailed)) for path in paths])

    def posts(self, status=None, only_valid=False):
        for path in self.dir(status=status):
            post = self.post(path)
            if only_valid and not post.is_valid():
                continue
            yield post

    def __getitem__(self, path):
        return self.post(path)

    def __len__(self):
        return len(self.dir())

    def __iter__(self):
        return self.posts()

    def __contains__(self, path):
        return self.has_post(path)

    # -------------- Post submission / addition user flow --------------------

    def add(self, kp, path=None, update=False, **kwargs):  # Create a new knowledge post draft
        assert isinstance(kp, KnowledgePost), "One can only add KnowledgePost objects to a KnowledgeRepository."
        path = path or kp.path
        if not path:
            raise ValueError("Post path not provided for Knowledge Post, and one is not specified within the knowledge post. Either add the path to post headers using `path: <path>` or specify the project path on the command line adding `-p <path>` to the current command.")
        path = self._kp_path(path)
        path = self.config.path_parse(path)

        current_datetime = datetime.datetime.now()
        authors = kp.headers['authors']
        new_authors = [self.config.username_parse(author) for author in authors]
        if new_authors != authors or kp.headers['updated_at'] < current_datetime:
            kp.update_headers(authors=new_authors, updated_at=current_datetime)

        for postprocessor in self.config.postprocessors:
            KnowledgePostProcessor._get_subclass_for(postprocessor).process(kp)

        cleanup_kwargs = self._add_prepare(kp, path, update, **kwargs)

        self._kp_save(kp, path, update=update)

        if cleanup_kwargs:
            kwargs.update(cleanup_kwargs)
        self._add_cleanup(kp, path, update, **kwargs)

        return kp

    @abstractmethod
    def _add_prepare(self, kp, path, update=False):
        raise NotImplementedError

    @abstractmethod
    def _add_cleanup(self, kp, path, update=False):
        raise NotImplementedError

    def revise(self, kp, path, **kwargs):
        return self.add(kp, path, update=True, **kwargs)

    def submit(self, path):  # Submit a post for review
        return self._submit(self._kp_path(path))

    @abstractmethod
    def _submit(self, path):  # Submit a post for review
        raise NotImplementedError

    def accept(self, path):  # Submit a post for review
        return self._accept(self._kp_path(path))

    @abstractmethod
    def _accept(self, path):  # Submit a post for review
        raise NotImplementedError

    def publish(self, path):  # Publish a post for general perusal
        return self._publish(self._kp_path(path))

    @abstractmethod
    def _publish(self, path):  # Publish a post for general perusal
        raise NotImplementedError

    def unpublish(self, path):  # Unpublish a post for general perusal
        return self._unpublish(self._kp_path(path))

    @abstractmethod
    def _unpublish(self, path):  # Unpublish a post for general perusal
        raise NotImplementedError

    def remove(self, path, all=False):
        return self._remove(self._kp_path(path), all=all)

    @abstractmethod
    def _remove(self, path, all=False):
        raise NotImplementedError

    # ----------- Knowledge Post Data Retrieval/Pushing Methods --------------------

    def _kp_repository_uri(self, path):
        return self.uri

    @abstractmethod
    def _kp_uuid(self, path):
        raise NotImplementedError

    def _kp_path(self, path, rel='/'):
        if path is None:
            return None
        path = os.path.relpath(os.path.abspath(os.path.join(rel, path)), rel)
        if os.name == 'nt':
            path = path.replace(os.path.sep, os.path.altsep)
        assert all([not segment.endswith('.kp') for segment in path.split(
            '/')[:-1]]), "The post path may not contain a directory named '*.kp'."
        if path == '.' or path.startswith('..'):
            raise ValueError("Provided path '{}' is outside of the knowledge repository.".format(path))
        if not path.endswith('.kp'):
            path += '.kp'
        return path

    @abstractmethod
    def _kp_exists(self, path, revision=None):
        raise NotImplementedError

    @abstractmethod
    def _kp_status(self, path, revision=None, detailed=False):
        raise NotImplementedError

    @abstractmethod
    def _kp_get_revision(self, path, status=None):
        raise NotImplementedError

    @abstractmethod
    def _kp_get_revisions(self, path):
        raise NotImplementedError

    @abstractmethod
    def _kp_read_ref(self, path, reference, revision=None):
        raise NotImplementedError

    @abstractmethod
    def _kp_dir(self, path, parent=None, revision=None):
        raise NotImplementedError

    @abstractmethod
    def _kp_has_ref(self, path, reference, revision=None):
        raise NotImplementedError

    @abstractmethod
    def _kp_diff(self, path, head, base):
        raise NotImplementedError

    @abstractmethod
    def _kp_write_ref(self, path, reference, data, uuid=None, revision=None):
        raise NotImplementedError

    @abstractmethod
    def _kp_new_revision(self, path, uuid=None):
        raise NotImplementedError

    def _kp_web_uri(self, path):
        return self.config.web_uri(path)

    def _kp_save(self, kp, path, update=False):
        if not update and self.has_post(path):
            raise ValueError("A knowledge post with the same path already exists. To update it, set the update flag.")
        kp.uuid = self._kp_uuid(path) or kp.uuid
        kp.path = path
        kp.revision = self._kp_new_revision(path, uuid=kp.uuid)
        kp.repository = self

        for ref in kp._dir():
            self._kp_write_ref(path, ref, kp._read_ref(ref), uuid=kp.uuid, revision=kp.revision)

    @property
    def web_uri(self):
        return self.config.web_uri()

    # ----------- Interface with web app ----------------------------------
    def get_app(self, *args, **kwargs):
        from . import app
        return self.config.prepare_app(app.KnowledgeFlask(self, *args, **kwargs))
